from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any

from debug_toolbar.middleware import DebugToolbarMiddleware
from debug_toolbar.utils import is_processable_html_response
from django.core.serializers.json import DjangoJSONEncoder
from django.http.request import MediaType
from django.template.loader import render_to_string
from django.utils.encoding import force_str

from undine.settings import undine_settings

if TYPE_CHECKING:
    from debug_toolbar.toolbar import DebugToolbar
    from django.http import HttpRequest, HttpResponse

__all__ = [
    "monkeypatch_middleware",
]


def monkeypatch_middleware() -> None:
    """Insert additional GraphiQL handling to the debug toolbar middleware."""
    original_postprocess = DebugToolbarMiddleware._postprocess  # noqa: SLF001

    def patched_postprocess(
        self: DebugToolbarMiddleware,
        request: HttpRequest,
        response: HttpResponse,
        toolbar: DebugToolbar,
    ) -> HttpResponse:
        response = original_postprocess(self, request, response, toolbar)
        if _is_graphql_view(request):
            handle_graphiql(request, response, toolbar)
        return response

    DebugToolbarMiddleware._postprocess = patched_postprocess  # noqa: SLF001


def handle_graphiql(request: HttpRequest, response: HttpResponse, toolbar: DebugToolbar) -> None:
    """Add debug toolbar data to GraphiQL view responses."""
    if is_processable_html_response(response):
        add_toolbar_update_script(response)
        return

    if not _is_json_response(response):
        return

    if _is_introspection_query(request):
        return

    add_debug_toolbar_data(response, toolbar)
    return


def add_toolbar_update_script(response: HttpResponse) -> None:
    """Add the JS script to the GraphiQL template for updating the toolbar after a request."""
    template = render_to_string("undine/graphiql_debug_toolbar_patch.html")
    response.write(template)
    if "Content-Length" in response:
        response["Content-Length"] = len(response.content)


def add_debug_toolbar_data(response: HttpResponse, toolbar: DebugToolbar) -> None:
    """
    Add data for the debug toolbar to the response.
    Will be used by '/template/undine/patch_debug_toolbar.html'.
    """
    content = force_str(response.content, encoding=response.charset)
    payload = json.loads(content)

    try:
        request_id = toolbar.request_id
    except AttributeError:
        #  Debug toolbar < 6.0.0 compatibility
        request_id = toolbar.store_id

    payload["debugToolbar"] = {"requestId": request_id, "panels": {}}

    for panel in reversed(toolbar.enabled_panels):
        payload["debugToolbar"]["panels"][panel.panel_id] = {
            "toolbarTitle": _call_if_callable(panel.nav_subtitle),
            "panelTitle": _call_if_callable(panel.title) if panel.has_content else None,
        }

    response.content = json.dumps(payload, cls=DjangoJSONEncoder)
    if "Content-Length" in response:
        response["Content-Length"] = len(response.content)


def _is_json_response(response: HttpResponse) -> bool:
    if getattr(response, "streaming", False):
        return False

    content_encoding = response.get("Content-Encoding", "")
    if content_encoding:
        return False

    content_type = response.get("Content-Type", "")
    if not content_type:
        return False

    media_type = MediaType(content_type)

    return media_type.match("application/json") or media_type.match("application/graphql-response+json")


def _is_introspection_query(request: HttpRequest) -> bool:
    try:
        body = json.loads(request.body)
    except Exception:  # noqa: BLE001
        return False

    return body.get("operationName") == "IntrospectionQuery"


def _is_graphql_view(request: HttpRequest) -> bool:
    if request.resolver_match is None:
        return False

    return request.resolver_match.view_name == f"undine:{undine_settings.GRAPHQL_VIEW_NAME}"


def _call_if_callable(obj: Any) -> Any:
    if callable(obj):
        return obj()
    return obj



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from channels.auth import AuthMiddlewareStack
from channels.consumer import AsyncConsumer
from channels.db import aclose_old_connections
from channels.exceptions import StopConsumer
from channels.routing import ProtocolTypeRouter, URLRouter
from channels.security.websocket import AllowedHostsOriginValidator
from django.urls import re_path

from undine.settings import undine_settings
from undine.utils.graphql.websocket import GraphQLOverWebSocketHandler

if TYPE_CHECKING:
    from asgiref.typing import WebSocketConnectEvent, WebSocketDisconnectEvent, WebSocketReceiveEvent
    from django.core.handlers.asgi import ASGIHandler


__all__ = [
    "GraphQLWebSocketConsumer",
    "get_websocket_enabled_app",
]


class GraphQLWebSocketConsumer(AsyncConsumer):
    """Channels consumer receiving messages from the WebSocket for the GraphQL over WebSocket Protocol."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.handler = GraphQLOverWebSocketHandler(websocket=self)

    async def websocket_connect(self, message: WebSocketConnectEvent) -> None:
        await self.handler.connect()

    async def websocket_receive(self, message: WebSocketReceiveEvent) -> None:
        await self.handler.receive(data=message["text"])

    async def websocket_disconnect(self, message: WebSocketDisconnectEvent) -> None:
        await self.handler.disconnect()
        await aclose_old_connections()
        raise StopConsumer


def get_websocket_enabled_app(django_application: ASGIHandler) -> Any:  # pragma: no cover
    """
    Create the default routing configuration for supporting GraphQL over WebSocket.

    >>> # asgi.py
    >>> import os
    >>>
    >>> from django.core.asgi import get_asgi_application
    >>>
    >>> os.environ.setdefault("DJANGO_SETTINGS_MODULE", "...")
    >>> django_application = get_asgi_application()
    >>>
    >>> # Needs to be imported after 'django_application' is created!
    >>> from undine.integrations.channels import get_websocket_enabled_app
    >>>
    >>> application = get_websocket_enabled_app(django_application)
    """
    websocket_urlpatterns = [re_path(undine_settings.WEBSOCKET_PATH, GraphQLWebSocketConsumer.as_asgi())]

    return ProtocolTypeRouter({
        "http": django_application,
        "websocket": AllowedHostsOriginValidator(AuthMiddlewareStack(URLRouter(websocket_urlpatterns))),
    })



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

from graphql import NoSchemaIntrospectionCustomRule, specified_rules

from undine.settings import undine_settings
from undine.typing import DjangoRequestProtocol

from .max_alias_count import MaxAliasCountRule
from .max_complexity_rule import MaxComplexityRule
from .max_directive_count import MaxDirectiveCountRule
from .one_of_input_object import OneOfInputObjectTypeRule, core_implements_one_of_directive
from .visibility_rule import VisibilityRule

if TYPE_CHECKING:
    from graphql import ASTValidationRule


__all__ = [
    "DjangoRequestProtocol",
    "get_validation_rules",
]


def get_validation_rules(*, inside_request: bool = False) -> tuple[type[ASTValidationRule], ...]:
    """Get the GraphQL validation rules based on project current settings."""
    visibility_enabled = inside_request and undine_settings.EXPERIMENTAL_VISIBILITY_CHECKS
    return tuple(
        itertools.chain(
            specified_rules,
            [MaxAliasCountRule, MaxDirectiveCountRule, MaxComplexityRule],
            [] if not visibility_enabled else [VisibilityRule],
            [] if core_implements_one_of_directive() else [OneOfInputObjectTypeRule],
            [] if undine_settings.ALLOW_INTROSPECTION_QUERIES else [NoSchemaIntrospectionCustomRule],
            undine_settings.ADDITIONAL_VALIDATION_RULES,
        )
    )



from __future__ import annotations

from django.urls import path

from undine.settings import undine_settings

from .views import graphql_view_async, graphql_view_sync

app_name = "undine"

urlpatterns = [
    path(
        undine_settings.GRAPHQL_PATH,
        graphql_view_async if undine_settings.ASYNC else graphql_view_sync,
        name=undine_settings.GRAPHQL_VIEW_NAME,
    ),
]



from __future__ import annotations

import dataclasses
import itertools
from abc import ABC, abstractmethod
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Any, Self

from django.db import transaction  # noqa: ICN003

from undine.exceptions import GraphQLAsyncAtomicMutationNotSupportedError
from undine.settings import undine_settings
from undine.utils.graphql.utils import get_operation_definition, is_atomic_mutation
from undine.utils.reflection import delegate_to_subgenerator

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator, Awaitable, Callable, Generator

    from graphql import DocumentNode, ExecutionResult, GraphQLFieldResolver

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import DjangoRequestProtocol, GQLInfo, T

__all__ = [
    "ExecutionLifecycleHookManager",
    "LifecycleHook",
    "LifecycleHookContext",
    "OperationLifecycleHookManager",
    "ParseLifecycleHookManager",
    "ValidationLifecycleHookManager",
]


@dataclasses.dataclass(kw_only=True)
class LifecycleHookContext:
    """Context passed to a lifecycle hook."""

    source: str
    """Source GraphQL document string."""

    document: DocumentNode | None
    """Parsed GraphQL document AST. Available after parsing is complete."""

    variables: dict[str, Any]
    """Variables passed to the GraphQL operation."""

    operation_name: str | None
    """Name of the GraphQL operation."""

    extensions: dict[str, Any]
    """GraphQL operation extensions received from the client."""

    request: DjangoRequestProtocol
    """Django request during which the GraphQL request is being executed."""

    result: ExecutionResult | Awaitable[ExecutionResult | AsyncIterator[ExecutionResult]] | None
    """Execution result of the GraphQL operation. Adding a result here will cause an early exit."""

    lifecycle_hooks: list[LifecycleHook] = dataclasses.field(init=False)
    """Lifecycle hooks for this operation."""

    def __post_init__(self) -> None:
        hooks = itertools.chain(undine_settings.LIFECYCLE_HOOKS, [AtomicMutationHook])
        self.lifecycle_hooks = [hook(context=self) for hook in hooks]

    @classmethod
    def from_graphql_params(cls, params: GraphQLHttpParams, request: DjangoRequestProtocol) -> Self:
        return cls(
            source=params.document,
            document=None,
            variables=params.variables,
            operation_name=params.operation_name,
            extensions=params.extensions,
            request=request,
            result=None,
        )


class LifecycleHook:
    """
    Base class for lifecycle hooks.

    Override methods to hook into the lifecycle of the GraphQL execution.
    Only overridden methods will be used.
    """

    def __init__(self, context: LifecycleHookContext) -> None:
        """
        Initialize a hook to use for an operation.

        :param context: The context for the hook.
        """
        self.context = context
        """Information on the GraphQL operation is being executed."""

    # Sync hooks.
    # Anything before the yield statement will be executed before the hooking point.
    # Anything after the yield statement will be executed after the hooking point.

    def on_operation(self) -> Generator[None, None, None]:
        """Hooking point that encompasses the entire GraphQL operation (parsing, validation, and execution)."""
        yield

    def on_parse(self) -> Generator[None, None, None]:
        """Hooking point that encompasses the parsing of the GraphQL document into an AST."""
        yield

    def on_validation(self) -> Generator[None, None, None]:
        """Hooking point that encompasses the GraphQL AST validation."""
        yield

    def on_execution(self) -> Generator[None, None, None]:
        """Hooking point that encompasses the execution of the GraphQL AST against the GraphQL schema."""
        yield

    # Resolver hook must be named 'resolve' to be compatible with 'MiddlewareManager'.
    # Should always call the resolver with the arguments as shown below.
    def resolve(self, resolver: GraphQLFieldResolver, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        """Hooking point that encompasses the resolution of each GraphQL field."""
        return resolver(root, info, **kwargs)

    # Async versions delegate to the sync versions by default

    async def on_operation_async(self) -> AsyncGenerator[None, None]:
        """Same as `on_operation`, but async. Use sync version by default."""
        with delegate_to_subgenerator(self.on_operation()) as gen:
            for _ in gen:
                yield

    async def on_parse_async(self) -> AsyncGenerator[None, None]:
        """Same as `on_parse`, but async. Use sync version by default."""
        with delegate_to_subgenerator(self.on_parse()) as gen:
            for _ in gen:
                yield

    async def on_validation_async(self) -> AsyncGenerator[None, None]:
        """Same as `on_validation`, but async. Use sync version by default."""
        with delegate_to_subgenerator(self.on_validation()) as gen:
            for _ in gen:
                yield

    async def on_execution_async(self) -> AsyncGenerator[None, None]:
        """Same as `on_execution`, but async. Use sync version by default."""
        with delegate_to_subgenerator(self.on_execution()) as gen:
            for _ in gen:
                yield


# Builtin hooks


class AtomicMutationHook(LifecycleHook):
    """
    Hook for executing multiple GraphQL mutations atomically
    if the `@atomic` directive is used on a mutation.
    """

    def __init__(self, context: LifecycleHookContext) -> None:
        super().__init__(context)

        self.is_atomic_mutation: bool = False
        self.error: BaseException | None = None

    def on_execution(self) -> Generator[None, None, None]:
        operation_definition = get_operation_definition(self.context.document, self.context.operation_name)
        self.is_atomic_mutation = is_atomic_mutation(operation_definition)

        if not self.is_atomic_mutation:
            yield
            return

        atomic = transaction.atomic()
        atomic.__enter__()  # noqa: PLC2801
        try:
            yield

        except BaseException as error:
            atomic.__exit__(error.__class__, error, error.__traceback__)
            raise

        else:
            if self.error is not None:
                atomic.__exit__(self.error.__class__, self.error, self.error.__traceback__)
            else:
                atomic.__exit__(None, None, None)

        finally:
            self.error = None

    async def on_execution_async(self) -> AsyncGenerator[None, None]:
        operation_definition = get_operation_definition(self.context.document, self.context.operation_name)
        self.is_atomic_mutation = is_atomic_mutation(operation_definition)

        if not self.is_atomic_mutation:
            yield
            return

        # `transaction.atomic` is not supported in async contexts.
        raise GraphQLAsyncAtomicMutationNotSupportedError

    def resolve(self, func: GraphQLFieldResolver, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        try:
            return func(root, info, **kwargs)

        except BaseException as error:
            # If an exception is thrown in a top-level resolver, it's likely that the mutation did not complete
            # correctly (e.g. a permission or a validation error) and we should rollback the transaction.
            if self.is_atomic_mutation and len(info.path.as_list()) == 1:
                self.error = error
            raise


# Hook managers


class BaseLifecycleHookManager(ExitStack, AsyncExitStack, ABC):
    """Allows executing multiple lifecycle hooks at once."""

    def __init__(self, *, hooks: list[LifecycleHook]) -> None:
        self.hooks = hooks
        super().__init__()

    @abstractmethod
    def enter_sync(self, hook: LifecycleHook) -> Callable[[], Generator[None, None, None]]: ...

    @abstractmethod
    def enter_async(self, hook: LifecycleHook) -> Callable[[], AsyncGenerator[None, None]]: ...

    def __enter__(self) -> Self:
        for hook in self.hooks:
            method = self.enter_sync(hook)
            if method is not None:
                self.enter_context(contextmanager(method)())
        return super().__enter__()

    async def __aenter__(self) -> Self:
        for hook in self.hooks:
            method = self.enter_async(hook)
            if method is not None:
                await self.enter_async_context(asynccontextmanager(method)())
        return await super().__aenter__()


class OperationLifecycleHookManager(BaseLifecycleHookManager):
    """Manager for lifecycle hooks for the whole operation."""

    def enter_sync(self, hook: LifecycleHook) -> Callable[[], Generator[None, None, None]] | None:
        if hook.__class__.on_operation == LifecycleHook.on_operation:
            return None
        return hook.on_operation

    def enter_async(self, hook: LifecycleHook) -> Callable[[], AsyncGenerator[None, None]] | None:
        if (
            hook.__class__.on_operation == LifecycleHook.on_operation
            and hook.__class__.on_operation_async == LifecycleHook.on_operation_async
        ):
            return None
        return hook.on_operation_async


class ParseLifecycleHookManager(BaseLifecycleHookManager):
    """Manager for lifecycle hooks in the parse step."""

    def enter_sync(self, hook: LifecycleHook) -> Callable[[], Generator[None, None, None]] | None:
        if hook.__class__.on_parse == LifecycleHook.on_parse:
            return None
        return hook.on_parse

    def enter_async(self, hook: LifecycleHook) -> Callable[[], AsyncGenerator[None, None]] | None:
        if (
            hook.__class__.on_parse == LifecycleHook.on_parse
            and hook.__class__.on_parse_async == LifecycleHook.on_parse_async
        ):
            return None
        return hook.on_parse_async


class ValidationLifecycleHookManager(BaseLifecycleHookManager):
    """Manager for lifecycle hooks in the validation step."""

    def enter_sync(self, hook: LifecycleHook) -> Callable[[], Generator[None, None, None]] | None:
        if hook.__class__.on_validation == LifecycleHook.on_validation:
            return None
        return hook.on_validation

    def enter_async(self, hook: LifecycleHook) -> Callable[[], AsyncGenerator[None, None]] | None:
        if (
            hook.__class__.on_validation == LifecycleHook.on_validation
            and hook.__class__.on_validation_async == LifecycleHook.on_validation_async
        ):
            return None
        return hook.on_validation_async


class ExecutionLifecycleHookManager(BaseLifecycleHookManager):
    """Manager for lifecycle hooks in the execution step."""

    def enter_sync(self, hook: LifecycleHook) -> Callable[[], Generator[None, None, None]] | None:
        if hook.__class__.on_execution == LifecycleHook.on_execution:
            return None
        return hook.on_execution

    def enter_async(self, hook: LifecycleHook) -> Callable[[], AsyncGenerator[None, None]] | None:
        if (
            hook.__class__.on_execution == LifecycleHook.on_execution
            and hook.__class__.on_execution_async == LifecycleHook.on_execution_async
        ):
            return None
        return hook.on_execution_async


# Decorators


def with_lifecycle_hooks_manager(
    manager: type[BaseLifecycleHookManager],
) -> Callable[[Callable[[LifecycleHookContext], T]], Callable[[LifecycleHookContext], T]]:
    def decorator(func: Callable[[LifecycleHookContext], T]) -> Callable[[LifecycleHookContext], T]:
        @wraps(func)
        def wrapper(context: LifecycleHookContext) -> T:
            with manager(hooks=context.lifecycle_hooks):
                return func(context)

        return wrapper

    return decorator


with_operation_lifecycle_hooks_manager = with_lifecycle_hooks_manager(OperationLifecycleHookManager)
with_parse_lifecycle_hooks_manager = with_lifecycle_hooks_manager(ParseLifecycleHookManager)
with_validation_lifecycle_hooks_manager = with_lifecycle_hooks_manager(ValidationLifecycleHookManager)
with_execution_lifecycle_hooks_manager = with_lifecycle_hooks_manager(ExecutionLifecycleHookManager)


def with_lifecycle_hooks_manager_async(
    manager: type[BaseLifecycleHookManager],
) -> Callable[[Callable[[LifecycleHookContext], Awaitable[T]]], Callable[[LifecycleHookContext], Awaitable[T]]]:
    def decorator(
        func: Callable[[LifecycleHookContext], Awaitable[T]],
    ) -> Callable[[LifecycleHookContext], Awaitable[T]]:
        @wraps(func)
        async def wrapper(context: LifecycleHookContext) -> T:
            async with manager(hooks=context.lifecycle_hooks):
                return await func(context)

        return wrapper

    return decorator


with_operation_lifecycle_hooks_manager_async = with_lifecycle_hooks_manager_async(OperationLifecycleHookManager)
with_parse_lifecycle_hooks_manager_async = with_lifecycle_hooks_manager_async(ParseLifecycleHookManager)
with_validation_lifecycle_hooks_manager_async = with_lifecycle_hooks_manager_async(ValidationLifecycleHookManager)
with_execution_lifecycle_hooks_manager_async = with_lifecycle_hooks_manager_async(ExecutionLifecycleHookManager)



from __future__ import annotations

from undine.utils.validators import validate_url

from ._definition import ScalarType

__all__ = [
    "GraphQLURL",
    "url_scalar",
]


url_scalar: ScalarType[str, str] = ScalarType(
    name="URL",
    description="Represents a valid URL.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc3986",
)

GraphQLURL = url_scalar.as_graphql_scalar()


@url_scalar.serialize.register
@url_scalar.parse.register
def _(value: str) -> str:
    if value:
        validate_url(value)
    return value



from __future__ import annotations

from decimal import Decimal, InvalidOperation
from inspect import cleandoc

from undine.exceptions import GraphQLScalarInvalidValueError

from ._definition import ScalarType

__all__ = [
    "GraphQLDecimal",
    "decimal_scalar",
]


decimal_scalar: ScalarType[Decimal, str] = ScalarType(
    name="Decimal",
    description=cleandoc(
        """
        Represents a number as a string for correctly rounded floating point arithmetic.
        Maps to the Python `decimal.Decimal` type.
        """
    ),
)

GraphQLDecimal = decimal_scalar.as_graphql_scalar()


@decimal_scalar.parse.register
def _(value: int) -> Decimal:
    return Decimal(value)


@decimal_scalar.parse.register
def _(value: str) -> Decimal:
    try:
        return Decimal(value)
    except InvalidOperation as error:
        raise GraphQLScalarInvalidValueError(typename="decimal") from error


@decimal_scalar.serialize.register
def _(value: int) -> str:
    return str(Decimal(value))


@decimal_scalar.serialize.register
def _(value: str) -> str:
    try:
        parsed_value = Decimal(value)
    except InvalidOperation as error:
        raise GraphQLScalarInvalidValueError(typename="decimal") from error

    return str(parsed_value)


@decimal_scalar.serialize.register
def _(value: Decimal) -> str:
    return str(value)



from __future__ import annotations

from .parse_annotations import parse_first_param_type, parse_parameters, parse_return_annotation
from .parse_docstring import docstring_parser, parse_class_attribute_docstrings
from .parse_graphql_params import GraphQLRequestParamsParser
from .parse_nullability import parse_is_nullable
from .parse_relation_info import parse_model_relation_info

__all__ = [
    "GraphQLRequestParamsParser",
    "docstring_parser",
    "parse_class_attribute_docstrings",
    "parse_first_param_type",
    "parse_is_nullable",
    "parse_model_relation_info",
    "parse_parameters",
    "parse_return_annotation",
]
